

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;
import java.util.regex.*;

/**
 * 
 * read data from an OWL document, analyze and transform them to a DB SQL
 * readable data
 *  
 */
public class Transformer {

    private Config cfgType; // the Owl-->DB type config file
    private Config cfgConn; // config file for the connection   
    private boolean exportToSQL;
    private boolean importToDBS;
    private String schemaFile;
    private String dataFile;
         

    /**
     * Default Constructor
     */
    public Transformer(String configFile) throws Exception {
        cfgConn = new Config(configFile);
        this.schemaFile = cfgConn.getProperty("schemaFile");
        this.dataFile = cfgConn.getProperty("dataFile");
        Database myDB = createVirtualDB(readDocument(schemaFile));        
        
        // export the SQL schema & data to files
        if ((cfgConn.getProperty("exportToSQL").toLowerCase()).equals("yes")) {
            exportToSQL = true;
            String schemaSQL = cfgConn.getProperty("schemaSQL");
            String dataSQL = cfgConn.getProperty("dataSQL");
            boolean sql_dropTables;
            if ((cfgConn.getProperty("sql_dropTables").toLowerCase()).equals("yes")) { sql_dropTables = true; } else { sql_dropTables = false;}
            
            // schema
            if ((cfgConn.getProperty("sql_exportSchema").toLowerCase()).equals("yes"))
                schemaToSQL(schemaSQL,myDB,sql_dropTables);	
        	// data
        	if ((cfgConn.getProperty("sql_exportData").toLowerCase()).equals("yes")) 
        	    dataToSQL(readDocument(dataFile),dataSQL);                                                
        }
        // export the SQL schema & data to DBS...yipppiii
        if ((cfgConn.getProperty("dbs_importToDBS").toLowerCase()).equals("yes")) {        	
        	boolean dbs_dropTables;
        	if ((cfgConn.getProperty("dbs_dropTables").toLowerCase()).equals("yes")) { dbs_dropTables = true; } else { dbs_dropTables = false;}        	        	
        	boolean dbs_importSchema;
        	if ((cfgConn.getProperty("dbs_importSchema").toLowerCase()).equals("yes")) { dbs_importSchema = true; } else { dbs_importSchema = false;}
        	boolean dbs_importData;
        	if ((cfgConn.getProperty("dbs_importData").toLowerCase()).equals("yes")) { dbs_importData = true; } else { dbs_importData = false;}
        	
        	exportToDB(myDB,dbs_dropTables,dbs_importSchema,dbs_importData);
        	//export the data to dbs ?
        	
        }      
            
    }    
    /**
     * write the SQL representation of a database db to a file
     * @param fileName : output file
     * @param db : source database
     */
    public void schemaToSQL(String fileName,Database db, boolean dropTables) {
        try{            
            System.out.println("[+] Exporting to SQL file : " + fileName + ".");
            FileWriter writer = new FileWriter(fileName);
            writer.write(db.toSQL(dropTables));
            writer.close();                                                           
        }catch (Exception ex){
            System.err.println("[-] Exporting to SQL file : " + fileName + "." + ex.getMessage());
        }        
    }
    /**
     * create the database db in a DBS
     * @param db : source database 
     */
    public void exportToDB(Database db, boolean dropTables, boolean importSchema, boolean importData) {        
        try {            
            System.out.println("[+] Exporting to DBS.");
            ConnectionDB connDB = new ConnectionDB(cfgConn);
            connDB.getConnection();
            if (importSchema)
                db.toDB(connDB,dropTables);
            //WARNING : ALPHA BETA version debug
            if (importData)
                dataToDB(readDocument(dataFile),connDB);
            connDB.Close();
        } catch (Exception ex) {                        
            System.err.println("[-] Exporting to DBS. " + ex.getMessage());            
        }
    }    
    /**
     * read a OWL document to a JDOM tree
     * @param fileName : input OWL file
     */
    public Document readDocument(String fileName) throws Exception {
        try {
            System.out.println("[+] Reading OWL document : " + fileName + ".");
            SAXBuilder builder = new SAXBuilder();
            builder.setValidation(false);
            Document doc = builder.build(fileName);
            return doc;
        } catch (IOException e) {
            System.err.println("[-] Reading OWL document : " + fileName + " is not existing.");
        } catch (JDOMException e) {
            System.err.println("[-] Reading OWL document : JDOM problem.");
        }
        return null;
    }
    /**
     * output a JDOM document to a file
     * @param doc : a JDOM document
     * @param saveFile : output file
     */
    public void outputDocumentToFile(Document doc, String saveFile)
            throws Exception {
        try {
            XMLOutputter outp = new XMLOutputter();
            FileWriter writer = new FileWriter(saveFile);
            outp.output(doc, writer);
            writer.close();
        } catch (java.io.IOException e) {
            System.err.println("[-] Outputting JDOM document to file : file not existing.");
        }
    }    
    /**
     * make a DB representation (in memory) from a OWL db source DB
     * representation includes : 1. Database 2. Tables 3. Columns 4. Primary
     * Keys 5. Type 6. Foreign Keys
     * @param doc : a JDOM document
     */
    public Database createVirtualDB(Document doc) {
        try {
            System.out.println("[+] Creating virtual DB.");
            Database myDB = new Database();
            Element rootElement = doc.getRootElement();
            List myClasses = rootElement.getChildren("Class", rootElement
                    .getNamespace("owl"));
            Iterator itr = myClasses.iterator();
            while (itr.hasNext()) {
                Element myClass = (Element) itr.next();
                String myClassID = myClass.getAttributeValue("ID", myClass
                        .getNamespace("rdf"));
                ///System.out.println("-Class : " + myClassID);
                //get the type : database, table...
                Element myType = myClass.getChild("type", rootElement
                        .getNamespace("rdf"));
                String myTypeResource = myType.getAttributeValue("resource",
                        rootElement.getNamespace("rdf"));
                ///System.out.println(" Type : " + myTypeResource);
                if (myTypeResource
                        .equals("http://www.laborda.org/RDF/relational.owl#Database"))
                        myDB.setName(myClassID);
                if (myTypeResource
                        .equals("http://www.laborda.org/RDF/relational.owl#Table")) {
                    Table tbl = new Table(myClassID);
                    myDB.addTable(tbl);
                    //get the colunms
                    List myColumns = myClass.getChildren("hasColumn",
                            rootElement.getNamespace("db"));
                    Iterator itrColumns = myColumns.iterator();
                    ///System.out.println("n column : " + myColumns.size());
                    while (itrColumns.hasNext()) {
                        Element myColumn = (Element) itrColumns.next();
                        String myColumnResource = myColumn.getAttributeValue(
                                "resource", rootElement.getNamespace("rdf"));
                        ///System.out.println(" hasColumn : " +
                        // myColumnResource);
                        ///System.out.println("[SQL:COLUMN] " +
                        // myColumnResource);
                        // myColumnResource is like : #TableName.ColumnName
                        String myColumnName = myColumnResource.replaceFirst("#"
                                + myClassID + ".", "");
                        Column col = new Column(myColumnName);
                        tbl.addColumn(col);
                    }
                    //get the PK
                    Element myIsIdentifiedBy = myClass.getChild(
                            "isIdentifiedBy", rootElement.getNamespace("db"));
                    if (myIsIdentifiedBy != null) {
                        Element myPrimaryKey = myIsIdentifiedBy.getChild(
                                "PrimaryKey", rootElement.getNamespace("db"));
                        List myPkeys = myPrimaryKey.getChildren("hasColumn",
                                rootElement.getNamespace("db"));
                        Iterator itrPk = myPkeys.iterator();
                        //System.out.println("n pk : " + myPkeys.size());
                        while (itrPk.hasNext()) {
                            Element myPkey = (Element) itrPk.next();
                            String myPkeyResource = myPkey
                                    .getAttributeValue("resource", rootElement
                                            .getNamespace("rdf"));
                            //System.out.println(myPkeyResource);
                            ///System.out.println("[SQL:PKEY ] " +
                            // myPkeyResource);
                            //get column and set the pk attribute
                            Column col = tbl.getColumn(myPkeyResource
                                    .replaceFirst("#" + myClassID + ".", ""));
                            if (col != null) col.setPrimaryKey(true);
                        }
                    }
                }
            }
            List myDatatypeProperties = rootElement.getChildren(
                    "DatatypeProperty", rootElement.getNamespace("owl"));
            Iterator itrDP = myDatatypeProperties.iterator();
            ///System.out.println(myDatatypeProperties.size());
            while (itrDP.hasNext()) {
                Element myDatatypeProperty = (Element) itrDP.next();
                String myDatatypePropertyID = myDatatypeProperty
                        .getAttributeValue("ID", myDatatypeProperty
                                .getNamespace("rdf"));
                //System.out.println("-DP : " + myDatatypePropertyID);
                Element myType2 = myDatatypeProperty.getChild("type",
                        rootElement.getNamespace("rdf"));
                String myTypeResource2 = myType2.getAttributeValue("resource",
                        rootElement.getNamespace("rdf"));
                if (myTypeResource2
                        .equals("http://www.laborda.org/RDF/relational.owl#Column")) {
                    Element myDomain = myDatatypeProperty.getChild("domain",
                            rootElement.getNamespace("rdfs"));
                    String myDomainResource = myDomain.getAttributeValue(
                            "resource", rootElement.getNamespace("rdf"));
                    //get the table associated with this element
                    Table tbl = myDB.getTable(myDomainResource.replaceAll("#",
                            ""));
                    String myTblName = tbl.getName();
                    ///System.out.println("tab : " +tbl.getName());
                    Column col = tbl.getColumn(myDatatypePropertyID
                            .replaceFirst(myTblName + ".", ""));
                    ///System.out.println("col : " + col.getName());
                    //get the length
                    Element myLength = myDatatypeProperty.getChild("length",
                            rootElement.getNamespace("db"));
                    if (myLength != null) {
                        String myLengthResource = myLength.getAttributeValue(
                                "resource", rootElement.getNamespace("rdf"));
                        //System.out.println(myLengthResource);
                        col.setSize(myLengthResource);
                    }
                    //get the scale
                    Element myScale = myDatatypeProperty.getChild("scale",
                            rootElement.getNamespace("db"));
                    if (myScale != null) {
                        String myScaleResource = myScale.getAttributeValue(
                                "resource", rootElement.getNamespace("rdf"));
                        //System.out.println(myScaleResource);
                        col.setScale(myScaleResource);
                    }
                    //get the type
                    Element myRange = myDatatypeProperty.getChild("range",
                            rootElement.getNamespace("rdfs"));
                    String myRangeResource = myRange.getAttributeValue(
                            "resource", rootElement.getNamespace("rdf"));
                    ///System.out.println(myRangeResource);
                    //removes the prefix "http://www.w3.org/2001/XMLSchema#"
                    //and translate the type from the OWL to the DB type
                    String myType = OwlToDbType(myRangeResource.replaceFirst(
                            "http://www.w3.org/2001/XMLSchema#", ""));
                    col.setType(myType);
                    //get the FK
                    Element myReferences = myDatatypeProperty.getChild(
                            "references", rootElement.getNamespace("db"));
                    if (myReferences != null) {
                        String myReferencesResource = myReferences
                                .getAttributeValue("resource", rootElement
                                        .getNamespace("rdf"));
                        //System.out.println(" FK
                        // :(source)"+myDatatypePropertyID+"
                        // (foreign)"+myReferencesResource);
                        List sourceList = Explode(myDatatypePropertyID, '.');
                        List foreignList = Explode(myReferencesResource
                                .replaceFirst("#", ""), '.');
                        String sourceTable = (String) sourceList.get(0);
                        String sourceColumn = (String) sourceList.get(1);
                        String foreignTable = (String) foreignList.get(0);
                        String foreignColumn = (String) foreignList.get(1);
                        //check if the tables exist
                        Table tabSource = myDB.getTable(sourceTable);
                        Table tabForeign = myDB.getTable(foreignTable);
                        Column colSource = tabSource.getColumn(sourceColumn);
                        Column colForeign = tabForeign.getColumn(foreignColumn);
                        //table exist ? yes
                        tabSource.addForeignKey(tabSource.getName(), colSource
                                .getName(), tabForeign.getName(), colForeign
                                .getName());
                        //System.out.println(fkey.getName());
                    }
                }
            }
            return myDB;
        } catch (Exception ex) {
            System.err.println("[-] Transformer.CreateVirtualDB.");
            return null;
        }
    }

    /**
     * return the DB type corresponding to the OWL type
     */
    public String OwlToDbType(String type) throws Exception {
        if (cfgType == null) cfgType = new Config("TypeMapping.properties");
        return cfgType.getProperty(type);
    }

    /**
     * Explode a token separated string, and return the list of elements
     * 
     * @param string : token separated string
     * @param separator : a separator
     * @return the list of elements
     */
    public List Explode(String string, char separator) throws Exception {
        List elements = new ArrayList(100);
        try {
            int indexFirstSeparator = string.indexOf(separator);
            int counter = 0;
            while (indexFirstSeparator != -1) {
                String mot = string.substring(0, indexFirstSeparator);
                elements.add(mot);
                string = string.substring(indexFirstSeparator + 1, string
                        .length());
                indexFirstSeparator = string.indexOf(separator);
                counter++;
            }
            //adds the last part of the string containing the last element
            elements.add(string);
        } catch (Exception e) {
            System.err.println("[-] Explode.");
        }
        return elements;
    }

    /***************************************************************************
     * DATABASE POPULATION
     */
    /**
     * Read an OWL document and generate the SQL code for the Database
     * Population.
     */
    public void dataToSQL(Document doc, String dataSQL) throws Exception {
        try {
            System.out.println("[+] Creating DB population SQL code : " + dataSQL + "." );
            StringBuffer result = new StringBuffer();
            result.append("--\n");
            result.append("-- DATABASE POPULATION\n");
            result.append("--\n");
            Element rootElement = doc.getRootElement();
            //get the list of the tables
            List myTables = rootElement.getChildren();
            Iterator itr = myTables.iterator();
            while (itr.hasNext()) {
                Element myTable = (Element) itr.next();
                result.append("INSERT INTO");
                result.append(" " + myTable.getName());
                //get the name and the value for each nuplet
                List myColumns = myTable.getChildren();
                //get the list of cols, removing the table prefix
                String myCols = " ("
                        + makeListName(myColumns, myTable.getName() + ".")
                        + ") ";
                String myValues = " (" + makeListValue(myColumns) + ")";
                result.append(myCols).append("VALUES").append(myValues);
                result.append("@\n");
                //System.out.println(result);
            }
            FileWriter writer = new FileWriter(dataSQL);
            writer.write(result.toString());
            writer.close();            
        } catch (Exception e) {
            System.err.println("[-] PopulateDB.");
        }
    }
    /**
     * Read an OWL document and import it to a DBS 
     */
    public void dataToDB(Document doc, ConnectionDB conn) throws Exception {
        try {
            System.out.println("[+] Exporting data population to DBS.");           
            Element rootElement = doc.getRootElement();
            List myTables = rootElement.getChildren();
            Iterator itr = myTables.iterator();
            while (itr.hasNext()) {
                StringBuffer result = new StringBuffer();
                Element myTable = (Element) itr.next();
                result.append("INSERT INTO");
                result.append(" " + myTable.getName());                
                List myColumns = myTable.getChildren();
                String myCols = " ("
                        + makeListName(myColumns, myTable.getName() + ".")
                        + ") ";
                String myValues = " (" + makeListValue(myColumns) + ")";
                result.append(myCols).append("VALUES").append(myValues);
                conn.Execute(result.toString());               
            }
        } catch (Exception e) {
            System.err.println("[-] Exporting data population to DBS.");
        }
    }

   
    /**
     * detect if the string is a SQL number (int, decimal...)
     * if it s a string -> return true
     */
    public boolean isString(String input){
        Pattern p = Pattern.compile("[^0-9.]");
        Matcher m = p.matcher(input);        
        boolean isString = m.find();        
        //if (isString)
        //    System.out.println("this is a string " + input + isString);
        //else 
        //    System.out.println("this is a nummer " + input + isString);        
        return isString;
    }
    
    /**
     * return a Name comma delimited list remove the prefix in case there is one
     * MODULTYP.BEZEICHNUNG --> BEZEICHNUNG
     */
    public String makeListName(List list, String Prefix) throws Exception {
        StringBuffer myList = new StringBuffer();
        String myElement = "";
        int size = list.size();
        Iterator itrList = list.iterator();
        myElement = ((Element) itrList.next()).getName().replaceFirst(Prefix,
                "");
        myList.append(myElement);
        while (itrList.hasNext()) {
            myList.append(",");
            myList.append(((Element) itrList.next()).getName().replaceFirst(
                    Prefix, ""));
        }
        return myList.toString();
    }

    /**
     * return a Value comma delimited list
     * for each value, check if it s a number (int, decimal,...) or string
     */
    public String makeListValue(List list) throws Exception {
        StringBuffer myList = new StringBuffer();
        int size = list.size();
        Iterator itrList = list.iterator();
        String value = ((Element) itrList.next()).getValue();
        
        if (isString(value))
            //it is a string, 'blablabla'
            myList.append("'").append(value).append("'");
        else 
            //it is a number, 12 or 1.2
            myList.append(value);
        
        while (itrList.hasNext()) {
            myList.append(",");
            value = ((Element) itrList.next()).getValue();
            if (isString(value))
                //it is a string, 'blablabla'
                myList.append("'").append(value).append("'");
            else 
                //it is a number, 12 or 1.2
                myList.append(value);                      
        }
        return myList.toString();
    }
    /* ************************************************************************************* */
}